<?php

namespace Ems\Services;

use Doctrine\ORM\EntityManager;
use Doctrine\DBAL\Connection;

use Ems\Services\MarcacaoPontoService;
use Ems\Util\Utils;

use Util\{ DateUtil, Environment, Repositories };

use Teknisa\Model\Gestaopessoas\Entities\GpeIntervaloshorario;

class HorarioService {

    /** @var EntityManager*/
    protected $entityManager;

    /** @var Connection */
    protected $connection;

    /** @var Environment */
    protected $environment;

    public function __construct(
        EntityManager $entityManager,
        Environment $environment
    ) {
        $this->entityManager = $entityManager;
        $this->connection  = $entityManager->getConnection();
        $this->environment = $environment;
    }
    
    public function getPrimeiroIntervaloHorario(
        int $nrHordiam
    ): ?GpeIntervaloshorario {
        return $this->entityManager->getRepository(Repositories::GPE_INTERVALOSHORARIO)->getPrimeiroIntervaloHorario(
            $this->environment->getNrorg(), $nrHordiam
        );
    }
    
    public function getUltimoIntervaloHorario(
        int $nrHordiam
    ): ?GpeIntervaloshorario {
        return $this->entityManager->getRepository(Repositories::GPE_INTERVALOSHORARIO)->getUltimoIntervaloHorario(
            $this->environment->getNrorg(), $nrHordiam
        );
    }
    
    public function horaEntradaIntervalo(
        GpeIntervaloshorario $intervalo
    ): float {
        return DateUtil::getHoraFloatDeString($intervalo->getHrentrada());
    }
    
    public function horaEntradaToleraAntesIntervalo(
        GpeIntervaloshorario $intervalo
    ): float {
        return DateUtil::getHoraFloatDeString($intervalo->getHrtoleraanteentrada());
    }
    
    public function horaSaidaIntervalo(
        GpeIntervaloshorario $intervalo
    ): float {
        return DateUtil::getHoraFloatDeString($intervalo->getHrsaida());
    }
    
    public function horaSaidaToleraAposIntervalo(
        GpeIntervaloshorario $intervalo
    ): float {
        return DateUtil::getHoraFloatDeString($intervalo->getHrtolerapostsaida());
    }
    
    public function getPrimeiraMarcacaoHorario(
        int         $nrHordiam,
        ?\DateTime  $data = null
    ): ?\DateTime {
        $data = !empty($data) ? $data : DateUtil::getDataAtual(true);
        $primeiroIntervalo = $this->getPrimeiroIntervaloHorario($nrHordiam);
        if (empty($primeiroIntervalo)) {
            return $data;
        }
        
        $horaEntrada = $this->horaEntradaToleraAntesIntervalo($primeiroIntervalo);
        $dataMarcacao = DateUtil::adicionaHorasAData($data, $horaEntrada);
        return MarcacaoService::marcacaoComDiaOcorrenciaReal($dataMarcacao, $primeiroIntervalo->getIddiaocorreentrada());
    }

    public function getPrimeiraHoraHorario(
        int         $nrHordiam,
        ?\DateTime  $data = null
    ): ?\DateTime {
        $data = !empty($data) ? $data : DateUtil::getDataAtual(true);
        $primeiroIntervalo = $this->getPrimeiroIntervaloHorario($nrHordiam);
        if (empty($primeiroIntervalo)) {
            return $data;
        }
        
        $horaEntrada = $this->horaEntradaToleraAntesIntervalo($primeiroIntervalo);
        return DateUtil::adicionaHorasAData($data, $horaEntrada);
    }
    
    public function getUltimaMarcacaoHorario(
        int         $nrHordiam,
        ?\DateTime  $data = null
    ): ?\DateTime {
        $data = !empty($data) ? $data : DateUtil::getDataAtual(true);
        $ultimoIntervalo = $this->getUltimoIntervaloHorario($nrHordiam);
        if (empty($ultimoIntervalo)) {
            return $data;
        }
        
        $horaSaida = $this->horaSaidaToleraAposIntervalo($ultimoIntervalo);
        $dataMarcacao = DateUtil::adicionaHorasAData($data, $horaSaida);
        return MarcacaoService::marcacaoComDiaOcorrenciaReal($dataMarcacao, $ultimoIntervalo->getIddiaocorresaida());
    }

    public function getUltimaHoraHorario(
        int         $nrHordiam,
        ?\DateTime  $data = null
    ): ?\DateTime {
        $data = !empty($data) ? $data : DateUtil::getDataAtual(true);
        $ultimoIntervalo = $this->getUltimoIntervaloHorario($nrHordiam);
        if (empty($ultimoIntervalo)) {
            return $data;
        }
        
        $horaSaida = $this->horaSaidaToleraAposIntervalo($ultimoIntervalo);
        return DateUtil::adicionaHorasAData($data, $horaSaida);
    }

    /**
     *  Rotina que verifica se os intervalos de horário não incluem marcações 'DA' ou 'FA' que remetem à dias anteriores
     * 
     *  @param $nrorg
     *  @param $nrhordiam
     *  @return boolean
     */
    public function horarioIncluiDiaAnterior($nrorg, $nrhordiam): bool {
        $intervalosHorario = $this->entityManager->getRepository(Repositories::GPE_INTERVALOSHORARIO)->findBy(
            ['nrorg' => $nrorg, 'nrhordiam' => $nrhordiam]
        );
        if (empty($intervalosHorario)) {
            return false;
        }
        
        foreach($intervalosHorario as $intervaloHorario) {
            if (in_array($intervaloHorario->getIddiaocorreentrada(), MarcacaoService::IDDIAOCORRENCIA_ANTERIORES) ||
                in_array($intervaloHorario->getIddiaocorresaida(), MarcacaoService::IDDIAOCORRENCIA_ANTERIORES)) {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     *  Rotina que verifica se os intervalos de horário não incluem marcações 'DS' ou 'FS' que remetem à dias seguintes
     * 
     *  @param $nrorg
     *  @param $nrhordiam
     *  @return boolean
     */
    public function horarioIncluiDiaSeguinte($nrorg, $nrhordiam): bool {
        $intervalosHorario = $this->entityManager->getRepository(Repositories::GPE_INTERVALOSHORARIO)->findBy(
            ['nrorg' => $nrorg, 'nrhordiam' => $nrhordiam]
        );
        if (empty($intervalosHorario)) {
            return false;
        }
        
        foreach($intervalosHorario as $intervaloHorario) {
            if (in_array($intervaloHorario->getIddiaocorreentrada(), MarcacaoService::IDDIAOCORRENCIA_SEGUINTES) ||
                in_array($intervaloHorario->getIddiaocorresaida(), MarcacaoService::IDDIAOCORRENCIA_SEGUINTES)) {
                return true;
            }
        }
        
        return false;
    }

    public function horarioIncluiDiaCorrente($nrorg, $nrhordiam): bool {
        $intervalosHorario = $this->entityManager->getRepository(Repositories::GPE_INTERVALOSHORARIO)->findBy(
            ['nrorg' => $nrorg, 'nrhordiam' => $nrhordiam]
        );
        if (empty($intervalosHorario)) {
            return false;
        }
        
        foreach($intervalosHorario as $intervaloHorario) {
            if (in_array($intervaloHorario->getIddiaocorreentrada(), MarcacaoService::IDDIAOCORRENCIA_CORRENTES) ||
                in_array($intervaloHorario->getIddiaocorresaida(), MarcacaoService::IDDIAOCORRENCIA_CORRENTES)) {
                return true;
            }
        }
        
        return false;
    }
}