<?php

namespace Ems\Services;

use Doctrine\ORM\EntityManager;
use Zeedhi\Framework\DependencyInjection\InstanceManager;

use Ems\Util\{ Utilities, Distancia, ExpectedException, Utils, Time };
use Ems\Repository\{
    MarcacaoRepository,
    PeriodoRepository,
    EstruturaRepository,
    EscalaRepository
};
use Ems\Services\{
    EstruturaService,
    VinculoService,
    ApuracaoService,
    ReportService,
    HorarioService,
    OperadorService
};

use Ems\Exception\CustomException;
use Ems\Services\MarcacaoPontoService;
use Ems\Util\ArrayUtil;

use Util\{ DateUtil, Environment, NovoCodigo, Messages, Repositories };

// MENSAGENS
define('messagesMarcacaoService', [
    'ERRO_LOCALIZACAO_INCORRETA'        => 'Você não está na área definida como de marcação do ponto! Diferença de @ metros.',
    'ERRO_REGIST_PONTO_EXIGE_FOTO'      => 'A foto é obrigatória no registro de ponto',
    'ERRO_PONTO_ACESSO_NAO_ENCONTRADO'  => 'A estrutura do vínculo não possui ponto de acesso cadastrado!',
    'ERRO_PERIODO_NAO_ENCONTRADO'       => 'Período não encontrado!',
    'ERRO_FILE_SERVER'                  => 'Não foi salvar a marcação com a foto. Log: @'
]);

class MarcacaoService {
    private const pontoAcessoDefaults = array(
        'VRDISTANCIAMAXREF' => 0,
        'IDOBRIGAFOTO' => 'N',
        'DSLATITUDE' => null,
        'DSLONGITUDE' => null
    );

    // ID's de alteração das marcaões apuradas
    const IDFIXAENTSAID_SAIDA     = 'S';
    const IDFIXAENTSAID_ENTRADA   = 'E';

    // Parametros de obtenção de marcações no dia
    const ORDEM_MARCACAO_ENTRADA    = Utils::ORDEM_ASC;
    const ORDEM_MARCACAO_SAIDA      = Utils::ORDEM_DESC;
    const POSICAO_MARCACAO_ENTRADA  = 1;
    const POSICAO_MARCACAO_SAIDA    = 1;

    // ID's Entrada/Saída Marcação
    const IDMARCACAO_ENTRADA    = 'E';
    const IDMARCACAO_SAIDA      = 'S';
    
    // Tipo de dia de ocorrência das marcações
    const IMPORTACAO_DEFAULT_DIA = self::IDDIAOCORRENCIA_FORCA_DIA_CORRENTE;
    const IDDIAOCORRENCIA_DIA_ANTERIOR          = 'DA';
    const IDDIAOCORRENCIA_DIA_CORRENTE          = 'DC';
    const IDDIAOCORRENCIA_DIA_SEGUINTE          = 'DS';
    const IDDIAOCORRENCIA_FORCA_DIA_ANTERIOR    = 'FA';
    const IDDIAOCORRENCIA_FORCA_DIA_CORRENTE    = 'FC';
    const IDDIAOCORRENCIA_FORCA_DIA_SEGUINTE    = 'FS';
    
    // Agrupações de dias de ocorrencia
    const IDDIAOCORRENCIA_ANTERIORES    = [self::IDDIAOCORRENCIA_DIA_ANTERIOR, self::IDDIAOCORRENCIA_FORCA_DIA_ANTERIOR];
    const IDDIAOCORRENCIA_CORRENTES     = [self::IDDIAOCORRENCIA_DIA_CORRENTE, self::IDDIAOCORRENCIA_FORCA_DIA_CORRENTE];
    const IDDIAOCORRENCIA_SEGUINTES     = [self::IDDIAOCORRENCIA_DIA_SEGUINTE, self::IDDIAOCORRENCIA_FORCA_DIA_SEGUINTE];
    const CADAUTOMARC_DEFAULT_IDDIAOCORRENCIA_DIA_CORRENTE = self::IDDIAOCORRENCIA_DIA_CORRENTE;
    
    protected const DEFAULT_DIA_ID = self::IDDIAOCORRENCIA_DIA_CORRENTE;
    
    // Controle de alteracao de marcaoes na manipulacao de marcacoes
    const UPDATEINTERVALOALMOCO_ALTERAMARCAOES_DEFAULT = true;
    const UPDATEINTERVALOALMOCO_ALTERAMARCAOES_SETMARCACAO = false;

    const IMPORTACAO_TEMPO_TOLERANCIA = '02:00';
    const IMPORTACAO_TEMPO_TOLERANCIA_ANTERIOR = self::IMPORTACAO_TEMPO_TOLERANCIA;
    const IMPORTACAO_TEMPO_TOLERANCIA_SEGUINTE = self::IMPORTACAO_TEMPO_TOLERANCIA;

    const IMPORTACAO_DIA_CORRENTE = self::IDDIAOCORRENCIA_FORCA_DIA_CORRENTE;
    const IMPORTACAO_DIA_SEGUINTE = self::IDDIAOCORRENCIA_FORCA_DIA_SEGUINTE;
    const IMPORTACAO_DIA_ANTERIOR = self::IDDIAOCORRENCIA_FORCA_DIA_ANTERIOR;

    /** @var Connection */
    protected $connection;

    /** @var ApuracaoService */
    protected $apuracaoService;

    /** @var VinculoService */
    protected $vinculoService;

    /** @var OcorrenciaService */
    protected $ocorrenciaService;

    /** @var ParamOrgService */
    protected $paramOrgService;

    /** @var HorarioService */
    protected $horarioService;

    /** @var Environment */
    protected $environment;

    /** @var MarcacaoRepository */
    protected $marcacaoRepository;

    /** @var EntityManager*/
    protected $entityManager;
    
    /** @var int[] */
    protected $ultimasMarcacoesCriadas;

    /** @var mixed[] */
    protected $marcacoesNaoRemovidasOnEdit;

    public function __construct(
        EntityManager           $entityManager,
        Environment             $environment,
        MarcacaoRepository      $marcacaoRepository,
        PeriodoRepository       $periodoRepository,
        EstruturaRepository     $estruturaRepository,
        EscalaRepository        $escalaRepository,
        Utilities               $utilities,
        MarcacaoPontoService    $marcacaoPontoService,
        EstruturaService        $estruturaService,
        OcorrenciaService       $ocorrenciaService,
        VinculoService          $vinculoService,
        ApuracaoService         $apuracaoService,
        ParamOrgService         $paramOrgService,
        ReportService           $reportService,
        HorarioService          $horarioService,
        OperadorService         $operadorService
    ) {
        $this->entityManager        = $entityManager;
        $this->environment          = $environment;
        $this->marcacaoRepository   = $marcacaoRepository;
        $this->periodoRepository    = $periodoRepository;
        $this->estruturaRepository  = $estruturaRepository;
        $this->escalaRepository     = $escalaRepository;
        $this->utilities            = $utilities;
        $this->marcacaoPontoService = $marcacaoPontoService;
        $this->estruturaService     = $estruturaService;
        $this->ocorrenciaService    = $ocorrenciaService;
        $this->connection           = $entityManager->getConnection();
        $this->vinculoService       = $vinculoService;
        $this->apuracaoService      = $apuracaoService;
        $this->paramOrgService      = $paramOrgService;
        $this->reportService        = $reportService;
        $this->horarioService       = $horarioService;
        $this->operadorService      = $operadorService;

        $this->marcacoesNaoRemovidasOnEdit = [];
    }

    /**
     * Função responsável pelo registro da marcação
     */
    public function setMarcacao(
        string  $nrvinculom,
        string  $nrorg,
        ?string $cdoperador,
        ?string $dslatitude,
        ?string $dslongitude,
        ?string $photoContent = null,
                $nrColetor = null,
        ?string $forcedDateTime = null
    ): array {
        try {
            $pontoAcessoValidado = $this->validaPontoAcesso($nrorg, $nrvinculom);

            $existeFotoRequisicao = Utils::isPhotoNotEmpty($photoContent);

            // Validação do local
            $localizacaoValida = $this->getValidacaoMarcacaoDistancia(
                $nrvinculom, $nrorg, $dslatitude,
                $dslongitude, $pontoAcessoValidado
            );

            if (!$localizacaoValida['DISTANCIAVALIDA']) {
                $mensagemException = str_replace('@', $localizacaoValida['DISTANCIADIFERENCA'], constant('messagesMarcacaoService')['ERRO_LOCALIZACAO_INCORRETA']);
                throw new CustomException($mensagemException, array('DISTANCIADIFERENCA' => $localizacaoValida['DISTANCIADIFERENCA']));
            }

            // Validação da foto
            $validacoesFotoMarcacao = $this->getValidacaoMarcacaoFoto(
                $nrvinculom, $nrorg,
                $existeFotoRequisicao, $pontoAcessoValidado
            );

            if (!$validacoesFotoMarcacao['OBRIGACAOCUMPRIDA']) {
                throw new \Exception(constant('messagesMarcacaoService')['ERRO_REGIST_PONTO_EXIGE_FOTO']);
            }

            $dspathfoto = null;
            $fileExtension = 'jpeg';
            if ($validacoesFotoMarcacao['ENVIAFOTO']) {
                // Faz o upload do arquivo para o fileServer, caso uma foto tenha sido recebida e se a
                // estrutura gerencial exige a foto
                try {
                    $dspathfoto = $this->utilities->saveFileOnFileServer($photoContent, $fileExtension, $nrorg);
                } catch (\Exception $e) {
                    throw new \Exception(str_replace('@', $e->getMessage(), constant('messagesMarcacaoService')['ERRO_FILE_SERVER']));
                }
            }

            // Monta o que vai ser inserido como a marcação
            $novaMarcacao = $this->novaMarcacao(
                $nrvinculom, $nrorg, $cdoperador,
                $dslatitude, $dslongitude, $dspathfoto,
                $nrColetor, $forcedDateTime
            );

            // Insere a marcação no banco
            $insertSuccess = $this->marcacaoRepository->insereMarcacoesPontoManual(
                $novaMarcacao['NRVINCULOM'], $novaMarcacao['DTHRMARCACAOPONTO'], $novaMarcacao['DTHRAJUSTADA'], $novaMarcacao['IDENQUADRAHORARIO'],
                $novaMarcacao['IDMARCDESPREZADA'], $novaMarcacao['IDDIAOCORRENCIA'], $novaMarcacao['DSJUSTIFICATIVA'], $novaMarcacao['NRORGINCLUSAO'],
                $novaMarcacao['NRCOLETOR'], $novaMarcacao['IDREFERINTERVALO'], $novaMarcacao['CDCNPJESTRUT'], $novaMarcacao['NMRAZSOCESTRUT'],
                $novaMarcacao['CDCAEPFESTRUT'], $novaMarcacao['CDCEIESTRUT'], $novaMarcacao['CDTICKETAUT'], $novaMarcacao['DSLATITUDE'],
                $novaMarcacao['DSLONGITUDE'], $novaMarcacao['DSPATH'], $novaMarcacao['IDATIVO'], $novaMarcacao['CDOPERINCLUSAO'],
                $novaMarcacao['NRORG']
            );
            
            if ($insertSuccess) {
                $dataMarcacao = substr($novaMarcacao['DTHRMARCACAOPONTO'], 0, 10);
                $this->updateIntervaloAlmocoPorTipoOcorrenciaTempoJornada(
                    $nrvinculom, $dataMarcacao, self::UPDATEINTERVALOALMOCO_ALTERAMARCAOES_SETMARCACAO
                );

                // Gera o relatorio comprovante
                $comprovanteFile = $this->getComprovanteReport($insertSuccess);

                // Busca pelas ultimas marcacoes registradas no dia
                $marcacoesDia = $this->getHorariosMarcacoesDia(
                    $nrvinculom,
                    \DateTime::createFromFormat(DateUtil::FORMATO_BRASILEIRO_DATAHORA, $novaMarcacao['DTHRAJUSTADA'])
                );
            }
            
            return [
                'success' => !empty($insertSuccess),
                'nrmarcacaoponto' => (int) $insertSuccess,
                'file' => $comprovanteFile ?? null,
                'marcacoesDia' => $marcacoesDia,
            ];
        } catch (\Exception $e) {
            throw $e;
        }
    }

    public function getValidacaoMarcacaoDistancia(
        string  $nrvinculom,
        string  $nrorg,
        ?string $dslatitude,
        ?string $dslongitude,
                $pontoAcesso = null
    ): array {
        $pontoAcessoValidado = $this->validaPontoAcesso($nrorg, $nrvinculom, $pontoAcesso);

        // Validações do ponto de acesso
        $marcacaoCerca = $pontoAcessoValidado['VRDISTANCIAMAXREF'];

        // Validação do local
        $latitudeReferencia = $pontoAcessoValidado['DSLATITUDE'];
        $longitudeReferencia = $pontoAcessoValidado['DSLONGITUDE'];

        // Valida se deve checar a distância entre o ponto de referência e o ponto de batida do ponto
        $distancia = 0;

        $validacaoNecessaria = $marcacaoCerca != 0 && !empty($dslatitude) && !empty($dslongitude);
        if ($validacaoNecessaria) {
            $distancia = Distancia::getDistanceBetweenLocations(
                $latitudeReferencia,
                $longitudeReferencia,
                $dslatitude,
                $dslongitude
            );
        }

        // Confirma se a marcação está sendo realizada em uma zona válida
        $distanciaValida = $distancia <= $marcacaoCerca;
        $diferencaDistancia = $distancia - $marcacaoCerca;
        $diferencaDistancia = round($diferencaDistancia, 3);

        if ($distanciaValida) {
            $diferencaDistancia = 0;
        }

        return [
            'DISTANCIAVALIDA'       => $distanciaValida,
            'DISTANCIADIFERENCA'    => $diferencaDistancia,
            'VRDISTANCIAMAXREF'     => $pontoAcessoValidado['VRDISTANCIAMAXREF'],
            'DSLATITUDE'            => $pontoAcessoValidado['DSLATITUDE'],
            'DSLONGITUDE'           => $pontoAcessoValidado['DSLONGITUDE']
        ];
    }

    public function getValidacaoMarcacaoFoto(
        string $nrvinculom,
        string $nrorg,
        bool $existeFotoRequisicao,
        ?array $pontoAcesso = null
    ) {
        $pontoAcessoValidado = $this->validaPontoAcesso($nrorg, $nrvinculom, $pontoAcesso);

        $idObrigaFoto = $pontoAcessoValidado['IDOBRIGAFOTO'];
        $estruturaExigeFotoMarcacao = $idObrigaFoto == 'S';

        $obrigacaoCumprida = false;
        $enviaFoto = false;

        if ($estruturaExigeFotoMarcacao && $existeFotoRequisicao) {
            $obrigacaoCumprida = true;
            $enviaFoto = true;
        } else if (!$estruturaExigeFotoMarcacao) {
            $obrigacaoCumprida = true;
            $enviaFoto = false;
        }

        return array(
            'ENVIAFOTO' => $enviaFoto,
            'OBRIGACAOCUMPRIDA' => $obrigacaoCumprida,
        );
    }

    public function novaMarcacao(
        string   $nrvinculom,
        string   $nrorg,
        ?string  $cdoperador,
        ?string  $dslatitude,
        ?string  $dslongitude,
                 $dspathfoto,
                 $nrColetor = null,
        ?string $forcedDateTime = null
    ) : array {
        $dataAtual = DateUtil::getDataAtual();
        if (!empty($forcedDateTime)) {
            $dataAtual = DateUtil::getDataDeString($forcedDateTime, DateUtil::FORMATO_DATAHORA, false);
        }

        if ($this->isNovaMaracaoDuplicante($nrvinculom, $dataAtual)) {
            throw new \Exception('Aguarde pelo menos três minutos para fazer um novo registro.');
        }
        
        $data = $dataAtual->format('d/m/Y');
        $horario = $dataAtual->format('H:i');

        if ($nrColetor === null) {
            $nrColetor = 24;
        }

        $entradaValidation = $this->verifyMarcacaoTime(
            $nrvinculom, $horario, $data, self::IDMARCACAO_ENTRADA
        );

        if (!$entradaValidation) {
            throw new \Exception(Messages::MARCACAO_LANC_ANTES_HORARIO_BASE);
        }

        // Hora marcação do ponto
        $dthoramarcacaoponto = $dataAtual->format('d/m/Y G:i:s');

        // Dados estrutura
        $dadosEmpresa = $this->estruturaRepository->getDadosEstruturaLegal($nrvinculom, $nrorg);

        $cdcnpjestrut = $dadosEmpresa['CDCNPJESTRUT'];
        $nmrazsocestrut = $dadosEmpresa['NMRAZSOCESTRUT'];
        $cdcaepfestrut = $dadosEmpresa['CDCAEPFESTRUT'];
        $cdceiestrut = $dadosEmpresa['CDCEIESTRUT'];

        // Ticket autenticação
        $ticketAut = $this->getMarcacaoTicketAut($nrvinculom, $nrorg, $dataAtual);

        // Obtem o dia de ocorrencia
        $idDiaOcorrencia = $this->getIdDiaMarcFromHora($nrvinculom, $dataAtual);

        $dadosMarcacao =  [
            "IDENQUADRAHORARIO" => "N",
            "IDDIAOCORRENCIA"   => $idDiaOcorrencia,
            "DSJUSTIFICATIVA"   => "",
            "NRCOLETOR"         => $nrColetor,
            "IDREFERINTERVALO"  => "",
            "CDCNPJESTRUT"      => $cdcnpjestrut,
            "NMRAZSOCESTRUT"    => $nmrazsocestrut,
            "CDCAEPFESTRUT"     => $cdcaepfestrut,
            "CDCEIESTRUT"       => $cdceiestrut,
            "CDTICKETAUT"       => $ticketAut,
            "DSLATITUDE"        => $dslatitude,
            "DSLONGITUDE"       => $dslongitude,
            "DSPATH"            => $dspathfoto,
            "IDATIVO"           => "S",
            "NRORGINCLUSAO"     => $nrorg,
            "CDOPERINCLUSAO"    => $cdoperador,
            "NRVINCULOM"        => $nrvinculom,
            "NRORG"             => $nrorg,
            "DTHRMARCACAOPONTO" => $dthoramarcacaoponto,
            "DTHRAJUSTADA"      => $dthoramarcacaoponto,
            "IDMARCDESPREZADA"  => 'N',
        ];

        return $dadosMarcacao;
    }

    public function getMarcacaoTicketAut($nrvinculom, $nrorg, $data = null) {
        if (empty($data)) {
            $data = new \DateTime();
        }

        if (!is_string($nrorg)) {
            $nrorg = strval($nrorg);
        }

        $nrorg = str_pad($nrorg, 4, "0", STR_PAD_LEFT);

        return $nrorg . $nrvinculom . $data->format('dmYGisu');
    }

    private function getPontoAcesso($nrorg, $nrvinculom) {
        $pontoAcesso = $this->estruturaRepository->getPontoAcesso($nrorg, $nrvinculom);

        if (empty($pontoAcesso)) {
            throw new \Exception(constant('messagesMarcacaoService')['ERRO_PONTO_ACESSO_NAO_ENCONTRADO']);
        }

        return $pontoAcesso;
    }

    private function validaPontoAcesso($nrorg, $nrvinculom, $pontoAcesso = null) {
        /* Tratativas exclusivas para o PG */
        if ($this->environment->isRequestFromPortalGestor()) {
            $exigeFotoPG = $this->paramOrgService->exigeFotoRegistroPontoPG();
            return [
                'VRDISTANCIAMAXREF' => self::pontoAcessoDefaults['VRDISTANCIAMAXREF'],
                'IDOBRIGAFOTO' => $exigeFotoPG ? 'S' : 'N',
                'DSLATITUDE' => self::pontoAcessoDefaults['DSLATITUDE'],
                'DSLONGITUDE' => self::pontoAcessoDefaults['DSLONGITUDE'],
            ];
        }

        /* Busca pelo ponto de acesso */
        
        if ($pontoAcesso == null) {
            $pontoAcesso = $this->getPontoAcesso($nrorg, $nrvinculom);
        }

        $vrdistanciamaxrefDefault = self::pontoAcessoDefaults['VRDISTANCIAMAXREF'];
        $idobrigafotoDefault = self::pontoAcessoDefaults['IDOBRIGAFOTO'];
        $dslatitudeDefault = self::pontoAcessoDefaults['DSLATITUDE'];
        $dslongitudeDefault = self::pontoAcessoDefaults['DSLONGITUDE'];

        $pontoAcesso['VRDISTANCIAMAXREF'] = !empty($pontoAcesso['VRDISTANCIAMAXREF']) ? $pontoAcesso['VRDISTANCIAMAXREF'] : $vrdistanciamaxrefDefault;
        $pontoAcesso['IDOBRIGAFOTO'] = !empty($pontoAcesso['IDOBRIGAFOTO']) ? $pontoAcesso['IDOBRIGAFOTO'] : $idobrigafotoDefault;
        $pontoAcesso['DSLATITUDE'] = !empty($pontoAcesso['DSLATITUDE']) ? $pontoAcesso['DSLATITUDE'] : $dslatitudeDefault;
        $pontoAcesso['DSLONGITUDE'] = !empty($pontoAcesso['DSLONGITUDE']) ? $pontoAcesso['DSLONGITUDE'] : $dslongitudeDefault;

        return $pontoAcesso;
    }

    public function getParesMarcacoesDia(
        int $nrVinculo,
        string $dia
    ) : array {

        $nrOrg = $this->environment->getNrOrg();

        $params = [
            'P_NRORG'           => $nrOrg,
            'P_NRVINCULOM'      => $nrVinculo,
            'P_DTAPURACAOPONTO' => $dia
        ];

        $paresMarcacoes = $this->marcacaoPontoService->getMarcacoesPonto($params);

        return $paresMarcacoes;
    }

    /**
     *  Rotina que permite remover marcações a partir do vínculo e número da marcação.
     *
     *  @param int     $nrVinculo,
     *  @param int     $nrMarcacao,
     *  @param string  $dia,
     *  @param string  $hrMarcacao,
     *  @param string  $idDia       | Default: self::DEFAULT_DIA_ID
     *
     *  @return $nrMarcacaoReturn
     *
     */
    public function removeMarcacao(
        int $nrVinculo,
        int $nrMarcacao
    ) : int {

        if (in_array($nrMarcacao, $this->marcacoesNaoRemovidasOnEdit)) {
            return 0;
        }

        $nrOrg = $this->environment->getNrOrg();

        $marcacao = $this->marcacaoRepository->
            getMarcacao($nrOrg, $nrMarcacao);


        // Se o coletor nao for aleravel sinaliza a marcacao para que
        // Uma possivel rotina de editar nao a inclua novamente
        if ($this->isColetorNotEditable($marcacao['NRCOLETOR'])) {
            array_push($this->marcacoesNaoRemovidasOnEdit, $nrMarcacao);
            return 0;
        }

        $date = $marcacao['DIA'];

        $almoco = $this->ocorrenciaService->getAlmoco($nrVinculo, $date);

        if ($almoco !== false) {
            $nrMarcacaoAlmoco = (int) $almoco['NRMARCACAOPONTO'];

            if ($nrMarcacaoAlmoco === $nrMarcacao) {
                $this->ocorrenciaService->removeAlmocoGestor($nrVinculo, $date);
            }
        }

        $this->ocorrenciaService->removeOcorrenciaByNrmarcacao($nrMarcacao);

        $params = [
            'P_NRORG'               => $nrOrg,
            'P_NRVINCULOM'          => $nrVinculo,
            'P_NRMARCACAOPONTO'     => $nrMarcacao
        ];

        $marcacoesRemovidas = $this->marcacaoRepository->removeMarcacao($params);

        return $marcacoesRemovidas;
    }

    public function removeParMarcacao(
        int     $nrVinculo,
        int     $nrMarcacaoEntrada,
        int     $nrMarcacaoSaida,
        bool    $atualizaInformacoesIntervaloAlmoco = true
    ) : int {
        
        // Validação de existência de marcações no mesmo horário
        if ($nrMarcacaoEntrada == $nrMarcacaoSaida) {
            throw new \Exception(Messages::MARCACAO_MARC_HOR_IGUAIS);
        }

        $nrOrg = $this->environment->getNrOrg();

        $nrMarcacoesRemovidas = array();

        // Obtem o dia das marcações inseridas
        $marcacao = $this->marcacaoRepository->getMarcacao($nrOrg, $nrMarcacaoEntrada) ?? $this->marcacaoRepository->getMarcacao($nrOrg, $nrMarcacaoSaida);
        if (empty($marcacao)) {
            return 0;
        }

        $date = $marcacao['DIA'];

        // Remove entrada
        $nrMarcacaoEntradaRemovida = $this->removeMarcacao($nrVinculo, $nrMarcacaoEntrada);

        // Remove saída
        $nrMarcacaoSaidaRemovida = $this->removeMarcacao($nrVinculo, $nrMarcacaoSaida);

        array_push($nrMarcacoesRemovidas, $nrMarcacaoEntradaRemovida);
        array_push($nrMarcacoesRemovidas, $nrMarcacaoSaidaRemovida);

        if ($atualizaInformacoesIntervaloAlmoco) {
            // Atualiza informações de marcações e ocorrência de almoço
            $this->updateIntervaloAlmocoPorTipoOcorrenciaTempoJornada($nrVinculo, $date);
        }

        return count($nrMarcacoesRemovidas);
    }

    /**
     *  Rotina que remove vários PARES DE MARCAÇÃO simultâneamente baseada na hora de ENTRADA/SAÍDA
     *
     *  @param  int     $nrVinculo,
     *  @param  string  $hrMarcacaoEntrada,
     *  @param  string  $hrMarcacaoSaida,
     *  @param  array   $dates
     */
    public function removeParesMarcacao(
        int     $nrVinculo,
        string  $hrMarcacaoEntrada,
        string  $hrMarcacaoSaida,
        array   $dates,
        string  $idEntrada = 'FC',
        string  $idSaida = 'FC'
    ) : array {

        try {

            $totalMarcacoesRemovidas = [];

            foreach ($dates as $date) {

                $marcacaoEntrada = $this->getMarcacaoByDiaMarcacao(
                    $nrVinculo,
                    $hrMarcacaoEntrada,
                    $date,
                    $idEntrada
                );

                $nrMarcacaoEntrada = (int) $marcacaoEntrada['NRMARCACAOPONTO'];

                $marcacaoSaida = $this->getMarcacaoByDiaMarcacao(
                    $nrVinculo,
                    $hrMarcacaoSaida,
                    $date,
                    $idSaida
                );

                $nrMarcacaoSaida = (int) $marcacaoSaida['NRMARCACAOPONTO'];

                $marcacoesRemovidas = $this->removeParMarcacao(
                    $nrVinculo,
                    $nrMarcacaoEntrada,
                    $nrMarcacaoSaida
                );

                $totalMarcacoesRemovidas[] = $marcacoesRemovidas;

            }

            return $totalMarcacoesRemovidas;

        } catch (\Exception $e) {
            throw $e;
        }

    }

    /**
     *  Rotina que permite salvar marcações a partir do vínculo, data e hora
     *
     *  @param int     $nrVinculo,
     *  @param int     $nrMarcacao,
     *  @param string  $dia,
     *  @param string  $hrMarcacao,
     *  @param string  $idDia       | Default: self::DEFAULT_DIA_ID
     *
     *  @return $nrMarcacaoReturn
     *
     */
    public function editMarcacao(
        int     $nrVinculo,
        int     $nrMarcacao,
        string  $dia,
        string  $hrMarcacao,
        string  $idDia      = self::DEFAULT_DIA_ID,
        $nrColetor = null
    ) : string {

        $marcacao = $this->marcacaoRepository->
            getMarcacao($this->environment->getNrorg(), $nrMarcacao);

        if (empty($marcacao) || $marcacao['HORA_MIN'] == $hrMarcacao) {
            return $nrMarcacao;
        }
        
        $this->removeMarcacao(
            $nrVinculo, $nrMarcacao);

        $nrMarcacao = $this->saveMarcacao(
            $nrVinculo, $dia, $hrMarcacao, $idDia, $nrColetor
        );

        return $nrMarcacao;
    }

    public function editParMarcacao(
        int $nrVinculo,
        int $nrMarcacaoEntrada,
        int $nrMarcacaoSaida,
        string $dia,
        string $hrEntrada,
        string $hrSaida,
        string $idDiaInicio = self::DEFAULT_DIA_ID,
        string $idDiaFim = self::DEFAULT_DIA_ID,
        $nrColetor = null
    ) : array {

        $this->connection->beginTransaction();
        $realizaValidacaoIntervaloAlmocoOnRemove = false;

        /* Sinaliza marcacoes nao alteradas */

        $marcacaoEntrada = $this->marcacaoRepository->
            getMarcacao($this->environment->getNrorg(), $nrMarcacaoEntrada);

        if (empty($marcacaoEntrada) || $marcacaoEntrada['HORA_MIN'] == $hrEntrada) {
            array_push($this->marcacoesNaoRemovidasOnEdit, $nrMarcacaoEntrada);
        }

        $marcacaoSaida = $this->marcacaoRepository->
            getMarcacao($this->environment->getNrorg(), $nrMarcacaoSaida);

        if (empty($marcacaoSaida) || $marcacaoSaida['HORA_MIN'] == $hrSaida) {
            array_push($this->marcacoesNaoRemovidasOnEdit, $nrMarcacaoSaida);
        }

        try {
            $this->removeParMarcacao(
                $nrVinculo, $nrMarcacaoEntrada, $nrMarcacaoSaida,
                $realizaValidacaoIntervaloAlmocoOnRemove
            );

            $nrMarcacoes = $this->saveParMarcacao(
                $nrVinculo, $dia, $hrEntrada,
                $hrSaida, $idDiaInicio, $idDiaFim, $nrColetor,
                null, null, null, $nrMarcacaoEntrada, $nrMarcacaoSaida
            );

            $this->connection->commit();
        } catch (\Exception $e) {
            $this->connection->rollBack();
            throw $e;
        }

        return $nrMarcacoes;
    }

    /**
     *  Rotina que edita um mesmo valor de ENTRADA/SAÍDA para diversas MARCAÇÕES simultâneamente
     *
     *  @parm   int     $nrVinculo,
     *  @param  string  $hrMarcacaoEntrada,
     *  @param  string  $hrMarcacaoSaida,
     *  @param  string  $novaHrEntrada,
     *  @param  string  $novaHrSaida,
     *  @param  array   $dates,
     *  @param  string  $idDiaInicio    = self::DEFAULT_DIA_ID,
     *  @param  string  $idDiaFim       = self::DEFAULT_DIA_ID
     */
    public function editParesMarcacao(
        int     $nrVinculo,
        string  $hrMarcacaoEntrada,
        string  $hrMarcacaoSaida,
        string  $novaHrEntrada,
        string  $novaHrSaida,
        array   $dates,
        string  $idDiaInicio    = self::DEFAULT_DIA_ID,
        string  $idDiaFim       = self::DEFAULT_DIA_ID,
        string  $novoIdInicio = null,
        string  $novoIdFim = null,
        $nrColetor = null
    ) : array {
        $novoIdInicio = $novoIdInicio ?? $idDiaInicio;
        $novoIdFim = $novoIdFim ?? $idDiaFim;

        $totalNrMarcacoesEditadas = [];

        foreach ($dates as $date) {

            $marcacaoEntrada = $this->getMarcacaoByDiaMarcacao(
                $nrVinculo,
                $hrMarcacaoEntrada,
                $date,
                $idDiaInicio
            );

            $nrMarcacaoEntrada = (int) $marcacaoEntrada['NRMARCACAOPONTO'];

            $marcacaoSaida = $this->getMarcacaoByDiaMarcacao(
                $nrVinculo,
                $hrMarcacaoSaida,
                $date,
                $idDiaFim
            );

            $nrMarcacaoSaida = (int) $marcacaoSaida['NRMARCACAOPONTO'];

            $nrMarcacoes = $this->editParMarcacao(
                $nrVinculo,
                $nrMarcacaoEntrada,
                $nrMarcacaoSaida,
                $date,
                $novaHrEntrada,
                $novaHrSaida,
                $novoIdInicio,
                $novoIdFim,
                $nrColetor
            );

            $totalNrMarcacoesEditadas[] = $nrMarcacoes;

        }

        return $totalNrMarcacoesEditadas;
    }

    /**
     *  Rotina que permite salvar marcações a partir do vínculo, data e hora
     *
     *  @param int     $nrVinculo,
     *  @param string  $date,
     *  @param string  $hora,
     *  @param string  $idDia       | Default: self::DEFAULT_DIA_ID
     *
     *  @return $nrMarcacaoReturn
     *
     */
    public function saveMarcacao(
        int      $nrVinculo,
        string   $date,
        ?string  $hora,
        string   $idDia  = self::DEFAULT_DIA_ID,
        ?string  $nrColetor = null,
        ?bool    $exigeAprovacao = false,
        ?int     $nrMarcacaoAnterior = null
    ) : ?string {
        if (empty($hora)) {
            return null;
        }
        
        if ($nrColetor === null) {
            $nrColetor = 24; // Marcação digitada
        }

        // Se estiver sinalizada como nao removida em alguma edicao anterior
        // A marcacao nao eh inserida
        if (in_array($nrMarcacaoAnterior, $this->marcacoesNaoRemovidasOnEdit)) {
            return null;
        }

        $nrOrg = (int) $this->environment->getNrOrg();
        $nrMarcacao = (int) NovoCodigo::geraCodigo($nrOrg, 'GPE_MARCACAOPONTO', 1, 22, false, false);

        $idDia = $this->fixDateId($idDia);

        $dateMarcacao = $this->dateIdToDate($date, $idDia);
        $dtHrMarcacao = "{$dateMarcacao} {$hora}:00";

        $params = [
            'P_NRORG'           => $nrOrg,
            'P_NRVINCULOM'      => $nrVinculo,
            'P_NRMARCACAO'      => $nrMarcacao,
            'P_DTHRMARCACAO'    => $dtHrMarcacao,
            'P_IDDIAOCORRENCIA' => $idDia,
            'P_NRCOLETOR'       => $nrColetor,
            'P_IDEXIGEAPROVACAO'=> ($exigeAprovacao ? Utils::SIM : Utils::NAO),
        ];

        $this->marcacaoRepository->saveMarcacao($params);

        $nrMarcacaoReturn = $params['P_NRMARCACAO'];

        return $nrMarcacaoReturn;
    }

    /**
     *  Rotina que permite salvar um par marcações a partir do vínculo, data e horas de entrada e saida
     *
     *  @param int     $nrVinculo,
     *  @param string  $date,
     *  @param string  $hrEntrada,
     *  @param ?string $hrSaida,
     *  @param string  $idDiaInicio         = self::DEFAULT_DIA_ID,
     *  @param string  $idDiaFim            = self::DEFAULT_DIA_ID,
     *  @param         $nrColetor           = null,
     *  @param ?string $justificativa       = null,
     *  @param ?int    $nrMotivoJust       = null
     *
     *  @return array
     *
     */
    public function saveParMarcacao(
        int      $nrVinculo,
        string   $date,
        string   $hrEntrada,
        ?string  $hrSaida,
        string   $idDiaInicio           = self::DEFAULT_DIA_ID,
        string   $idDiaFim              = self::DEFAULT_DIA_ID,
                 $nrColetor             = null,
        ?int     $nrMotivoJust          = null,
        ?string  $justificativa         = null,
        ?string  $anexoJustificativa    = null,
        ?int     $nrMarcacaoAnteriorEntrada = null,
        ?int     $nrMarcacaoAnteriorSaida   = null
    ): array {
        $exigeAprovacao = $this->environment->isRequestFromPortalFuncionario() && $this->paramOrgService->isAprovacaoMarcNecessaria();
        $horarioBase = $this->vinculoService->getHorarioBase($nrVinculo, $date, VinculoService::GETHORARIOBASE_THROW_EXCEPTION);
        
        // Um intervalo completo e necessario caso nao exija aprovacao
        if (!$exigeAprovacao && (empty($hrEntrada) || empty($hrSaida))) {
            throw new \Exception(Messages::MARCACAO_START_END_REQUIRED);
        }

        $exceedsTime = $this->verifyParExceedsTrabalho(
            $nrVinculo, $date, $hrEntrada,
            $hrSaida, $horarioBase
        );

        $entradaValidation = $this->verifyMarcacaoTime(
            $nrVinculo, $hrEntrada, $date, self::IDMARCACAO_ENTRADA, $horarioBase
        );

        $saidaValidation = $this->verifyMarcacaoTime(
            $nrVinculo, $hrSaida, $date, self::IDMARCACAO_SAIDA, $horarioBase
        );

        $this->verifyMarcacaoIsDuplicada(
            $nrVinculo, $date, $hrEntrada, $hrSaida
        );

        if (!$entradaValidation) {
            throw new \Exception(Messages::MARCACAO_LANC_ANTES_HORARIO_BASE);
        }

        if ($exceedsTime && !$saidaValidation) {
            throw new \Exception(Messages::MARCACAO_LANC_DEPOIS_TEMPO_SUPERIOR_HORARIO_BASE);
        }

        $idDiaInicio = $this->fixDateId($idDiaInicio);
        $idDiaFim    = $this->fixDateId($idDiaFim);

        // Insere entrada
        $nrMarcacaoEntradaInserida = $this->saveMarcacao(
            $nrVinculo, $date, $hrEntrada, $idDiaInicio, $nrColetor,
            $exigeAprovacao, $nrMarcacaoAnteriorEntrada
        );

        // Insere saída
        $nrMarcacaoSaidaInserida = $this->saveMarcacao(
            $nrVinculo, $date, $hrSaida, $idDiaFim, $nrColetor,
            $exigeAprovacao, $nrMarcacaoAnteriorSaida
        );

        $this->ultimasMarcacoesCriadas = array_filter(
            [$nrMarcacaoEntradaInserida, $nrMarcacaoSaidaInserida]
        );
        
        if ($exigeAprovacao) {
            $this->createOcorAprovUltMarcacoes(
                $nrVinculo,
                $date,
                $nrMotivoJust,
                $justificativa,
                $anexoJustificativa
            );
        }

        // Validação pra alterar intervalo de almoço. Alteração é necessária pois ao alterar uma de marcação, o intervalo de almoço pode estar sendo alterado.
        // O intervalo de almoço sendo alterado, a ocorrência de almoço deve ser alterada conforme o tempo de trabalho
        $this->updateIntervaloAlmocoPorTipoOcorrenciaTempoJornada($nrVinculo, $date);

        return $this->ultimasMarcacoesCriadas;
    }

    public function saveParesMarcacao(
        int     $nrVinculo,
        array   $dates,
        string  $hrEntrada,
        string  $hrSaida,
        string  $idDiaInicio        = self::DEFAULT_DIA_ID,
        string  $idDiaFim           = self::DEFAULT_DIA_ID,
        $nrColetor                  = null,
        ?int    $nrMotivoJust       = null,
        ?string $justificativa      = null,
        ?string $anexoJustificativa = null
    ) : array {

        $nrMaracaoesSalvos = [];

        try {
            $this->entityManager->beginTransaction();

            foreach ($dates as $date) {

                $nrMaracaoes = $this->saveParMarcacao(
                    $nrVinculo, $date, $hrEntrada,
                    $hrSaida, $idDiaInicio, $idDiaFim,
                    $nrColetor, $nrMotivoJust, $justificativa,
                    $anexoJustificativa
                );

                if (empty($nrMaracaoes)) {
                    throw new \Exception(Messages::MARCACAO_MULT_IMPOSSIVEL_SALVAR_MARCACAOES);
                }

                $nrMaracaoesSalvos = array_merge($nrMaracaoesSalvos, $nrMaracaoes);

            }

            $this->entityManager->commit();
            return $nrMaracaoesSalvos;
        } catch (\Exception $e) {
            $this->entityManager->rollback();
            throw $e;
        }

    }

    public static function dateIdToDate(
        string $referenceDate,
        string $dateId
    ): string {
        $date = $referenceDate;

        if ($dateId === 'FA' || $dateId === 'DA') {
            $date = Utils::addDaysToDate($referenceDate, 1);
        } else if ($dateId === 'FS' || $dateId === 'DS') {
            $date = Utils::subDaysFromDate($referenceDate, 1);
        }

        return $date;
    }

    public function getMarcacoesFotos(
        int $nrVinculo,
        string $date
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $marcacoes = $this->getMarcacoesDia($nrVinculo, $date);

        $marcacoesWithFile = array_filter($marcacoes, function($marcacao) {
            $hasFile =
                isset($marcacao['DSPATHFOTO']) &&
                $marcacao['DSPATHFOTO'] !== '';

            return $hasFile;
        });

        // Arruma os index do array [2 => 'a', 5 => 'b'] => [0 => 'a', 1 => 'b']
        $marcacoesWithFile = array_values($marcacoesWithFile);

        $filesData = array_map(function($marcacao) {
            $path = $marcacao['DSPATHFOTO'];
            $file = $this->utilities->downloadFile($path);

            $base64 = $file['base64'];
            $type = $file['type'];
            $horaMinuto = $marcacao['HORA_MINUTO'];
            $nrMarcacao = $marcacao['NRMARCACAOPONTO'];

            $fileData = [
                'NRMARCACAOPONTO' => $nrMarcacao,
                'type' => $type,
                'HORA_MINUTO' => $horaMinuto,
                'base64' => $base64
            ];

            return $fileData;
        }, $marcacoesWithFile);

        return $filesData;
    }

    public function getMarcacoesDia(
        int $nrVinculo,
        string $date
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'     => $nrOrg,
            'P_NRVINCULO' => $nrVinculo,
            'P_DATE'      => $date
        ];

        $marcacoes = $this->marcacaoRepository->getMarcacoesDia($params);

        return $marcacoes;
    }

    public static function fixDateId(string $dateId): string {
        if ($dateId === 'DA') {
            return 'FA';
        }

        if ($dateId === 'DS') {
            return 'FS';
        }

        if ($dateId === 'DC') {
            return 'FC';
        }

        return $dateId;
    }

    public function getMarcacaoByDiaMarcacao(
                $nrvinculom,
        string  $horaMarcacao,
        string  $diaMarcacao,
        string  $idMarcacao = 'FC'
    ) {
        $nrorg = $this->environment->getNrorg();

        $marcacao = $this->marcacaoRepository->getMarcacaoByDataMarcacao(
            $nrorg,
            $nrvinculom,
            $horaMarcacao,
            $diaMarcacao,
            false,
            $idMarcacao
        );

        if (empty($marcacao)) {
            throw new \Exception(Messages::MARCACAO_NOT_FOUND);
        }

        return $marcacao;
    }

    public function fixaEntrada($nrVinculom, $dia, $nrMarcacaoPonto) {
        $nrOrg = $this->environment->getNrorg();

        $this->marcacaoRepository->limpaIdFixaMarcacaoDiaEntrada(
            $nrOrg,
            $nrVinculom,
            $dia
        );

        $this->marcacaoRepository->setIdFixaMarcacaoEntrada(
            $nrOrg,
            $nrVinculom,
            $dia,
            $nrMarcacaoPonto
        );
    }

    public function fixaSaida($nrVinculom, $dia, $nrMarcacaoPonto) {
        $nrOrg = $this->environment->getNrorg();

        $this->marcacaoRepository->limpaIdFixaMarcacaoDiaSaida(
            $nrOrg,
            $nrVinculom,
            $dia
        );

        $this->marcacaoRepository->setIdFixaMarcacaoSaida(
            $nrOrg,
            $nrVinculom,
            $dia,
            $nrMarcacaoPonto
        );

    }

    /**
     *  Rotina que permite fixar a entrada e saída para marcações de dias determinados. O processo de fixar marcações consiste em alterar o campo IDFIXAENTSAID para 'E'
     * em caso de ENTRADA fixada e 'S' em caso de SAÍDA fixada.
     *
     *  @param $nrVinculom
     *  @param $dia
     *  @param $nrMarcacaoPontoFixaEntrada
     *  @param $nrMarcacaoPontoFixaSaida
     *
     *  @return bool    $alteracoesRealizadas
     *
     */
    public function fixaEntradaSaida(
        int    $nrVinculo,
        string $dia,
        int    $nrMarcacaoPontoFixaEntrada,
        int    $nrMarcacaoPontoFixaSaida
    ): array {
        $status = [
            'success' => true,
            'message' => null
        ];

        $apurado = $this->apuracaoService->verificaDiaTemResumo(
            $nrVinculo, $dia
        );

        if (!$apurado) {
            $message = 'Não é possível fixar entrada/saída em dias não apurados!';

            $status['success'] = false;
            $status['message'] = $message;

            return $status;
        }

        if (!empty($nrMarcacaoPontoFixaEntrada)) {
            $this->fixaEntrada($nrVinculo, $dia, $nrMarcacaoPontoFixaEntrada);
        }

        if (!empty($nrMarcacaoPontoFixaSaida)) {
            $this->fixaSaida($nrVinculo, $dia, $nrMarcacaoPontoFixaSaida);
        }

        return $status;
    }

    /**
     *  Rotina que permite alterar apenas a entrada e saída de dias determinados. As marcações de entrada e saída são fixadas respectivamente como entrada e saída.
     *
     *  @param          $nrVinculom
     *  @param Array    $marcacoes
     *      Format Expected:
     *          $marcacao['DIA']
     *          $marcacao['ENTRADA']
     *          $marcacao['SAIDA']
     *          $marcacao['ALMOCO']
     *
     *  @return bool    $marcacoesAlteradasComSucesso
     *
     */
    public function alteraEntradasSaidasFixadosComProjecao(
        $nrVinculom,
        $cdOperador,
        Array $marcacoes,
        $nrColetor = null
    ) {
        $marcacoesAlteradasComSucesso = true;

        foreach ($marcacoes as $marcacao) {

            $dia                    = $marcacao['DIA'];
            $hrFixaEntrada          = $marcacao['ENTRADA'];
            $hrFixaSaida            = $marcacao['SAIDA'];
            $tempoProjetarAlmoco    = $marcacao['ALMOCO'];

            $result = $this->alteraEntradaSaidaFixadosComProjecao(
                $nrVinculom,
                $cdOperador,
                $dia,
                $hrFixaEntrada,
                $hrFixaSaida,
                $tempoProjetarAlmoco,
                $nrColetor
            );

            $marcacoesAlteradasComSucesso = $marcacoesAlteradasComSucesso && $result;

        }

        return $marcacoesAlteradasComSucesso;
    }

    public function alteraEntradaSaidaFixadosComProjecao(
                $nrVinculom,
        string  $cdOperador,
        string  $dia,
        string  $hrEntrada,
        string  $hrSaida,
        string  $tempoProjetarAlmoco,
        $nrColetor = null
    ) {
        $this->entityManager->beginTransaction();

        try {
            $this->removeMarcacoesDia($nrVinculom, $dia);

            // Altera as marcações de entrada e saída
            $marcacoesInseridas = $this->alteraEntradaSaida($nrVinculom, $dia, $hrEntrada, $hrSaida, $nrColetor);

            // if (empty($marcacoesInseridas) || !isset($marcacoesInseridas['ENTRADA']) || !isset($marcacoesInseridas['SAIDA'])) {
            //     return false;
            // }

            // $nrMarcacaoEntrada  = $marcacoesInseridas['ENTRADA'];
            // $nrMarcacaoSaida    = $marcacoesInseridas['SAIDA'];

            // Fixa marcações

            // Fixar desativado momentaneamente
            // $this->fixaEntradaSaida($nrVinculom, $dia, $nrMarcacaoEntrada, $nrMarcacaoSaida);

            // Projetar intervalo almoco
            $tempoProjetarAlmocoFloat = DateUtil::formataHoraString($tempoProjetarAlmoco);

            $this->projetaMarcacoesIntervaloAlmoco(
                $nrVinculom, $cdOperador, $dia,
                $tempoProjetarAlmocoFloat, $hrEntrada, $hrSaida, $nrColetor
            );

            $this->entityManager->commit();
        } catch (\Exception $e) {
            $this->entityManager->rollback();
            throw $e;
        }

        return true;
    }

    public function alteraEntrada(
        $nrVinculom,
        $dia,
        $hrEntrada,
        $nrColetor = null
    ) {
        $nrOrg = $this->environment->getNrorg();

        // Update marcação
        // $nrMarcacao = $marcacaoEntrada['NRMARCACAOPONTO'];
        // return $this->editMarcacao(
        //     $nrVinculom,
        //     $nrMarcacao,
        //     $dia,
        //     $hrEntrada
        // );

        return $this->saveMarcacao(
            $nrVinculom,
            $dia,
            $hrEntrada,
            self::DEFAULT_DIA_ID,
            $nrColetor
        );
    }

    public function alteraSaida(
        $nrVinculom,
        $dia,
        $hrSaida,
        $nrColetor
    ) {
        $nrOrg = $this->environment->getNrorg();

        // Update marcação
        // $nrMarcacao = $marcacaoSaida['NRMARCACAOPONTO'];
        // return $this->editMarcacao(
        //     $nrVinculom,
        //     $nrMarcacao,
        //     $dia,
        //     $hrSaida
        // );

        return $this->saveMarcacao(
            $nrVinculom,
            $dia,
            $hrSaida,
            self::DEFAULT_DIA_ID,
            $nrColetor
        );
    }

     /**
     *  Rotina que permite alterar as marcações de entrada e de saída de um determinado dia
     *
     *  @param $nrVinculom
     *  @param $dia
     *  @param $hrEntrada
     *  @param $hrSaida
     *
     *  @return bool    $alteracoesRealizadasComSucesso
     *
     */
    public function alteraEntradaSaida(
        $nrVinculom,
        $dia,
        $hrEntrada,
        $hrSaida,
        $nrColetor = null
    ) {
        $marcacoesInseridas = array();

        if (!empty($hrEntrada)) {
            $nrMarcacaoEntrada = $this->alteraEntrada($nrVinculom, $dia, $hrEntrada, $nrColetor);
            $marcacoesInseridas['ENTRADA'] = (int) $nrMarcacaoEntrada;
        }

        if (!empty($hrSaida)) {
            $nrMarcacaoSaida = $this->alteraSaida($nrVinculom, $dia, $hrSaida, $nrColetor);
            $marcacoesInseridas['SAIDA'] = (int) $nrMarcacaoSaida;
        }

        $alteracoesRealizadasComSucesso = count($marcacoesInseridas) === 0 ? false : $marcacoesInseridas;
        return $alteracoesRealizadasComSucesso;

    }

    public function projetaMarcacoesIntervaloAlmoco(
        $nrVinculom,
        $cdOperador,
        $dia,
        $tempoIntervalo = null,
        $entrada = null,
        $saida = null,
        $nrColetor = null
    ) {
        //$this->removeMarcacoesIntervaloAlmoco($nrVinculom, $dia);

        $nrOrg = (int) $this->environment->getNrorg();
        $escalaTrabalho = $this->escalaRepository->getEscalaTrabalhoVinculo($nrOrg, $nrVinculom);

        $isEstagiario =
            !$escalaTrabalho['SEGENTRADA'] ||
            !$escalaTrabalho['SEGSAIDA'];

        if (empty($escalaTrabalho)) {
            throw new \Exception('Não foi encontrada escala de trabalho.');
        }

        $marcacoesInseridasProjecao = $this->insereMarcacoesProjecaoAlmoco(
            $nrVinculom, $dia, $escalaTrabalho,
            $tempoIntervalo, $entrada, $saida, $nrColetor
        );

        $nrMarcacaoEntradaProjecao = $marcacoesInseridasProjecao['ENTRADA'];

        if (!$isEstagiario) {
            $this->ocorrenciaService->insertAlmocoGestor($cdOperador, $nrVinculom, $dia, $nrMarcacaoEntradaProjecao);
        }

        return $marcacoesInseridasProjecao;
    }

    public function insereMarcacoesProjecaoAlmoco(
        $nrVinculom,
        $dia,
        $escalaTrabalho,
        $tempoIntervalo = null,
        $entrada = null,
        $saida = null,
        $nrColetor = null
    ) {
        $inicioIntervalo = '';
        $fimIntervalo = '';

        $estagiario =
            !$escalaTrabalho['SEGENTRADA'] ||
            !$escalaTrabalho['SEGSAIDA'];

        if ($estagiario) {
            $duracaoIntervaloTime = Time::fromFloat($tempoIntervalo);

            $entradaTime = Time::fromString($entrada);
            $saidaTime = Time::fromString($saida);

            $totalTrabTime = $saidaTime->sub($entradaTime);
            $metadeTime = $totalTrabTime->divide(2);

            $inicioIntervaloTime = $entradaTime->sum($metadeTime);
            $inicioIntervalo = $inicioIntervaloTime->toString();

            $fimIntervaloTime = $inicioIntervaloTime->sum($duracaoIntervaloTime);
            $fimIntervalo = $fimIntervaloTime->toString();
        } else {
            $inicioIntervalo = $escalaTrabalho['PRISAIDA'];

            // $tempoProjecaoIntervalo = is_null($tempoIntervalo) ? $escalaTrabalho['QTHRLIMITESEPHEXTRA'] : $tempoIntervalo;
            $fimIntervalo = DateUtil::somaValorHoraString($inicioIntervalo, $tempoIntervalo);
        }

        $this->verifyFimIntervalo($fimIntervalo, $saida);

        // Insere marcação início do intervalo
        $nrMarcacaoInicio = $this->saveMarcacao(
            $nrVinculom,
            $dia,
            $inicioIntervalo,
            self::DEFAULT_DIA_ID,
            $nrColetor
        );

        // Insere marcação fim do intervalo
        $nrMarcacaoFim = $this->saveMarcacao(
            $nrVinculom,
            $dia,
            $fimIntervalo,
            self::DEFAULT_DIA_ID,
            $nrColetor
        );

        $marcacoesInseridas = array('ENTRADA' => $nrMarcacaoInicio, 'SAIDA' => $nrMarcacaoFim);

        return $marcacoesInseridas;
    }

    public function verifyFimIntervalo(
        string $hrFimIntervalo,
        string $hrSaida
    ): bool {
        $fimIntervaloTime = Time::fromString($hrFimIntervalo);
        $saidaTime = Time::fromString($hrSaida);

        $comparison = $saidaTime->compare($fimIntervaloTime);

        if ($comparison === Time::LESS_THAN) {
            $message = 'O fim do intervalo não pode ser após a saída!';
            throw new ExpectedException($message);
        }

        return true;
    }

    public function removeMarcacoesIntervaloAlmoco($nrVinculom, $dia) {

        $nrOrg = $this->environment->getNrOrg();
        $parMarcacoesAlmoco = $this->marcacaoRepository->getParMarcacaoAlmoco($nrOrg, $nrVinculom, $dia);

        if (empty($parMarcacoesAlmoco)) {
            return false;
        }

        $nrMarcacaoEntrada  = (int) $parMarcacoesAlmoco['NRMARCACAOPONTO_ALMOCO'];
        $nrMarcacaoSaida    = (int) $parMarcacoesAlmoco['NRMARCACAOPONTO_POS_ALMOCO'];

        // Remove primeira marcação
        $this->removeMarcacao($nrVinculom, $nrMarcacaoEntrada);

        // Remove segunda marcação
        $this->removeMarcacao($nrVinculom, $nrMarcacaoSaida);

        $this->ocorrenciaService->removeAlmocoGestor($nrVinculom, $dia);

    }

    public function createBaliza(
        int $nrVinculo,
        string $date,
        string $time,
        string $idDia = 'FC',
        $nrColetor = null
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();
        $nrMarcacoes = [];
        $dateTime = Utils::getSqlDateTimeString($date, $time); // "DD/MM/YYYY HH:MI:SS"

        if ($nrColetor === null) {
            $nrColetor = 24; // Marcação digitada.
        }

        $this->connection->beginTransaction();

        try {
            for ($i = 0; $i < 2; $i++) {
                // O commit da NovoCodigo atrapalha as transações!!
                $nrMarcacao = NovoCodigo::geraCodigo(
                    $nrOrg, 'GPE_MARCACAOPONTO', 1, 20, true, false
                );

                array_push($nrMarcacoes, $nrMarcacao);

                $this->marcacaoRepository->insereMarcacoesPontoManual(
                    $nrVinculo, $dateTime, $dateTime, 'N', 'N', $idDia,
                    NULL, $nrOrg, $nrColetor, 'S', NULL, NULL, NULL, NULL, NULL,
                    NULL, NULL, NULL, 'S', 1, $nrOrg, $nrMarcacao
                );
            }

            $this->connection->commit();
        } catch (\Exception $e) {
            $this->connection->rollBack();
            throw $e;
        }

        return $nrMarcacoes;
    }

    /**
     *  Rotina dedicada à verificação da hora das marcações. Para isso, é feita uma sinalização se a marcação é de entrada (self::IDMARCACAO_ENTRADA) ou saída (self::IDMARCACAO_ENTRADA).
     * De maneira similar ao que é realizado no portalColaborador, entrada ou saída são comparadas com a primeira entrada/última saída definida na escala de trabalho considerando a tolerância definida
     * na nos parametros da organização.
     *
     * @param   int     $nrVinculo,
     * @param   ?string $timeMarcacao,
     * @param   string  $date,
     * @param   string  $idMarcacao,
     * @param           $horarioBase = null
     *
     * @return  bool
     */
    public function verifyMarcacaoTime(
        int     $nrVinculo,
        ?string $timeMarcacao,
        string  $date,
        string  $idMarcacao,
                $horarioBase = null
    ): bool {
        if (empty($timeMarcacao)) {
            return true;    
        }
        
        // Valida se na parametrização da organização, a opção para validação do horário base está ativada
        if (!$this->paramOrgService->paramOrgValidaHoraBasePortal()) {
            return true;
        }

        $idMarcacaoValidos = [self::IDMARCACAO_ENTRADA, self::IDMARCACAO_SAIDA];
        if (!in_array($idMarcacao, $idMarcacaoValidos)) {
            throw new \Exception('Id marcação inválido');
        }

        $horarioBase =
            $horarioBase ??
            $this->vinculoService->getHorarioBase($nrVinculo, $date, VinculoService::GETHORARIOBASE_THROW_EXCEPTION);

        if (empty($horarioBase)) {
            return true;
        }

        // Caso o dia em questão seja um dia de folga da escala ou um feeriado, a verificação não deve ser realizada
        if (in_array($horarioBase['IDTIPOHORARIO'], OcorrenciaService::IDTIPOHORARIO_FOLGA_COMP) || $this->vinculoService->isDiaFeriadoVinculo($nrVinculo, $date)) {
            return true;
        }

        // Valida para entrada
        if ($idMarcacao == self::IDMARCACAO_ENTRADA) {
            $entrada = (string) $horarioBase['PRIENTRADA'];

            // Não coloque o "(string)" nas linhas de baixo pois podem ser NULL
            $tolerancia = $horarioBase['HRTEMPOTOLERANCIAANT'];

            $timeMarcacaoTime = Time::fromString($timeMarcacao);
            $entradaTime = Time::fromString($entrada);

            if ($tolerancia !== null) {
                $toleranciaTime = Time::fromString($tolerancia);

                $entradaMenosToleranciaTime = $entradaTime->sub($toleranciaTime);

                $comparison = $timeMarcacaoTime->
                    compare($entradaMenosToleranciaTime);

                if ($comparison === Time::LESS_THAN) {
                    return false;
                }
            }
        } else {
            // Valida para saída
            $saida = $horarioBase['SEGSAIDA'] ?? $horarioBase['PRISAIDA'];

            // Não coloque o "(string)" nas linhas de baixo pois podem ser NULL
            $tolerancia = $horarioBase['HRTEMPOTOLERANCIAPOS'];

            $timeMarcacaoTime = Time::fromString($timeMarcacao);
            $saidaTime = Time::fromString($saida);

            if ($tolerancia !== null) {
                $toleranciaTime = Time::fromString($tolerancia);

                $saidaMaisToleranciaTime = $saidaTime->sum($toleranciaTime);

                $comparison = $timeMarcacaoTime->
                    compare($saidaMaisToleranciaTime);

                if ($comparison === Time::GREATER_THAN) {
                    return false;
                }
            }
        }

        return true;
    }

    /**
     *  Rotina que verifica se a inclusao de um novo intervalo ultrapassara o tempo limite para o dia
     *
     *  @param int     $nrVinculo,
     *  @param string  $date,
     *  @param string  $hrEntrada,
     *  @param ?string $hrSaida,
     *  @param         $horarioBase = null
     *
     *  @return bool
     */
    public function verifyParExceedsTrabalho(
        int     $nrVinculo,
        string  $date,
        string  $hrEntrada,
        ?string $hrSaida,
                $horarioBase = null
    ): bool {
        if (empty($hrSaida)) {
            return false;
        }
        
        $nrOrg = $this->environment->getNrOrg();

        $horarioBase =
            $horarioBase ??
            $this->vinculoService->getHorarioBase($nrVinculo, $date, VinculoService::GETHORARIOBASE_NOT_THROW_EXCEPTION);

        if (empty($horarioBase)) {
            return false;
        }

        // Caso o dia em questão seja um dia de folga da escala ou um feeriado, a verificação não deve ser realizada
        if ($horarioBase['IDTIPOHORARIO'] === OcorrenciaService::IDTIPOHORARIO_FOLGA || $this->vinculoService->isDiaFeriadoVinculo($nrVinculo, $date)) {
            return false;
        }

        $trabalho = $horarioBase['QTHORASTRABALHO'];
        $trabalhoTime = Time::fromString($trabalho);

        // Trata caso de existência de saldos a serem compensados por estagiários. Se houver débito a ser compensado, o tempo da jornada é extendido até o tempo que abata a diferença
        if ($this->vinculoService->getTipoVinculo($nrVinculo, $nrOrg) == VinculoService::TIPO_VINC_EST) {
            $dadosApuracao = $this->apuracaoService->getApuracaoPorData(
                $nrVinculo,
                $date
            );

            if (empty($dadosApuracao)) {
                throw new \Exception(Messages::APUR_DADOS_NOT_FOUND);
            }

            $saldoAnterior = isset($dadosApuracao['SALDO_BH_ANTERIOR']) ? ((float) $dadosApuracao['SALDO_BH_ANTERIOR']) * -1 : 0;
            $saldoAnteriorTime = Time::fromFloat($saldoAnterior);
            $trabalhoTime = $saldoAnteriorTime->sum($trabalhoTime);
        }

        $entradaTime = Time::fromString($hrEntrada);
        $saidaTime   = Time::fromString($hrSaida);

        $durationParTime = $saidaTime->sub($entradaTime);

        $horasTrabalhadas = $this->calculateHorasTrabalhadas($nrVinculo, $date);
        $horasTrabalhadasTime = Time::fromString($horasTrabalhadas);

        $totalTrabalhadoTime = $horasTrabalhadasTime->sum($durationParTime);

        $comparison = $totalTrabalhadoTime->compare($trabalhoTime);
        $exceeds = $comparison === Time::GREATER_THAN;

        return $exceeds;
    }

    public function calculateHorasTrabalhadas(
        int    $nrVinculo,
        string $date
    ): string {
        $nrOrg = (int) $this->environment->getNrOrg();

        $pares = $this->getParesMarcacoesDia($nrVinculo, $date);
        $trabalhadoTime = new Time(0);

        foreach ($pares as $par) {
            $trab = '0:00';

           $consideraTempo =
            $par['IDMARCACAO'] === 'Trabalho' ||
            $par['NRPROGOCORRENCIA_JUSTIFICADA'] !== '';

            if ($consideraTempo) {
                $trab = $par['TEMPO'];
            }

            $trabTime = Time::fromString($trab);
            $trabalhadoTime = $trabalhadoTime->sum($trabTime);
        }

        $trabalhado = $trabalhadoTime->toString();

        return $trabalhado;
    }

    /**
     *  Rotina responsável por calcular o a diferença entre o tempo trabalhado e o tempo definido na escala do vínculo.
     *
     *  @param  int    $nrVinculo
     *  @param  string $date
     *
     *  @return Time    $diffTempo
     *
     */
    public function calculateDiferencaHorasDia(
        int    $nrVinculo,
        string $date
    ): Time {
        $tempoTrabalhado    = $this->calculateHorasTrabalhadas($nrVinculo, $date);
        $tempoPrevisto      = $this->vinculoService->getHorarioBase($nrVinculo, $date, VinculoService::GETHORARIOBASE_NOT_THROW_EXCEPTION);

        if (empty($tempoTrabalhado) || empty($tempoPrevisto)) {
            return false;
        }

        $tempoTrabalhado    = Time::fromString($tempoTrabalhado);
        $tempoPrevisto      = Time::fromString($tempoPrevisto['QTHORASTRABALHO']);

        $diffTempo = $tempoTrabalhado->diff($tempoPrevisto);

        return $diffTempo;
    }

    public function getMarcacoesPontoDiaOpcoesFixarMarcacao(
        int    $nrVinculo,
        string $date
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'           => $nrOrg,
            'P_NRVINCULOM'      => $nrVinculo,
            'P_DTAPURACAOPONTO' => $date
        ];

        $marcacoes = $this->marcacaoRepository->
            getMarcacoesPontoDiaOpcoesFixarMarcacao($params);

        return $marcacoes;
    }

    public function reverterEntradaSaida(
        int    $nrVinculo,
        string $date
    ): bool {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'       => $nrOrg,
            'P_NRVINCULOM'  => $nrVinculo,
            'P_DATE'        => $date
        ];

        $this->marcacaoRepository->reverterEntradaSaida($params);

        return true;
    }

    /**
     *  Rotina responsável por validar a alteração para que as marcações e ocorrência de almoço se adequem à jornadas específicas. Em uma jornada com horas trabalhadas entre 04:00h e 06:00, o intervalo de
     * 15 minutos é obrigatório. Por causa disso, as marcações podem ser alteradas para que a jornada se adeque à essa questão. A ocorrência de almoço pode sofrer alteração no seu tipo (NRTIPOPROGOCORRENCIA) e
     * horas projetadas de almoço HRPROJETARINTERVALONPROG.
     *
     *  @param int       $nrVinculo
     *  @param string    $dia
     *  @param bool      $alterMarcacoesAlmoco. Default: UPDATEINTERVALOALMOCO_ALTERAMARCAOES_DEFAULT
     *
     *  @return void
     *
     */
    public function updateIntervaloAlmocoPorTipoOcorrenciaTempoJornada(
        int     $nrVinculo,
        string  $date,
        bool    $alterMarcacoesAlmoco=self::UPDATEINTERVALOALMOCO_ALTERAMARCAOES_DEFAULT
    ) {
        $defaultReturn  = OcorrenciaService::NRTIPOPROGOCORRENCIA_ALMOCO;
        $nrOrg          = $this->environment->getNrOrg();

        $parMarcacoesAlmoco = $this->marcacaoRepository->getParMarcacaoAlmoco($nrOrg, $nrVinculo, $date);
        if (empty($parMarcacoesAlmoco)) {
            return;
        }

        $nrEntradaAlmoco    = (int) $parMarcacoesAlmoco['NRMARCACAOPONTO_ALMOCO'];
        $nrSaidaAlmoco      = (int) $parMarcacoesAlmoco['NRMARCACAOPONTO_POS_ALMOCO'];

        // Modifica as marcações de almoço, caso necessário
        if ($alterMarcacoesAlmoco) {
            $this->updateMarcacoesIntervaloAlmoco(
                $nrVinculo,
                $date
            );
        }

        // Atualiza ocorrência de almoco, dependendo da jornada
        $this->ocorrenciaService->updateOcorrenciaAlmoco(
            $nrVinculo,
            $nrEntradaAlmoco,
            $date
        );
    }

    /**
     *  Rotina responsável por modificar marcações em caso de jornada com intervalo de 15 minutos. Em uma jornada com horas trabalhadas entre 04:00h e 06:00, o intervalo de 15 minutos é obrigatório.
     * Por causa disso, as marcações podem ser alteradas para que a jornada se adeque à essa questão.
     *
     *  @param int       $nrVinculo
     *  @param string    $dia
     *
     *  @return void
     *
     */
    public function updateMarcacoesIntervaloAlmoco(
        int     $nrVinculo,
        string  $dia
    ) {
        $nrOrg = $this->environment->getNrorg();
        
        // Validação condicionada ao parâmetro da organização "Validar Lançamento com Hora Base"
        if (!$this->paramOrgService->paramOrgValidaHoraBasePortal()) {
            return;
        }
        
        // Validação de apenas intervalos programados removida. O intervalo mínimo para qualquer intervalo com ocorrência de almoco é 00:15h.

        // $tipoOcorrenciaJornada = $this->ocorrenciaService->getTipoOcorrenciaAlmocoPorJornada($nrVinculo, $dia);
        // // Verifica se o tipo é de intervalo programado. Para esse tipo, o intervalo mínimo de 15 minutos tem que ser respeitado
        // $updateNaoNecessario = $tipoOcorrenciaJornada !== OcorrenciaService::NRTIPOPROGOCORRENCIA_ALMOCO_INT_NAO_PROG;
        // if ($updateNaoNecessario) {
        //     return;
        // }

        // Verifica se já existe um intervalo de almoço
        $parMarcacoesAlmoco = $this->marcacaoRepository->getParMarcacaoAlmoco($nrOrg, $nrVinculo, $dia);
        if (empty($parMarcacoesAlmoco)) {
            return;
        }
        
        // Tempo mínimo a ser projetado caso necessário. Conversão para segundos para comparação com o intervalo
        $tempoProjecao  = ApuracaoService::TEMPO_MINIMO_INTERVALO * (3600);
        
        $nrSaidaEntradaAlmoco = (int) $parMarcacoesAlmoco['NRMARCACAOPONTO_POS_ALMOCO'];
        
        // Conversão das datas-horas da marcação para \DateTime
        $entradaAlmoco  = DateUtil::getDataDeString($parMarcacoesAlmoco['DTHRMARCACAOPONTO_ALMOCO'], DateUtil::FORMATO_DATAHORA);
        $saidaAlmoco    = DateUtil::getDataDeString($parMarcacoesAlmoco['DTHRMARCACAOPONTO_POS_ALMOCO'], DateUtil::FORMATO_DATAHORA);
        
        // Cálculo da diferença entre os dias
        $tempoIntervalo = DateUtil::getIntervaloEmSegundos($entradaAlmoco, $saidaAlmoco);
        
        // Verifica se a projeção é necessária
        if ($tempoProjecao <= $tempoIntervalo) {
            return;
        }

        // Obtem nova saída de almoco projetada
        $novaSaidaAlmoco = DateUtil::adicionaIntervalo($entradaAlmoco, $tempoProjecao, DateUtil::SEGUNDOS);
        
        $dateTimeDia = DateUtil::getDataDeString($dia, DateUtil::FORMATO_DATA, true);
        $tempoLimiteDiaAnterior = !empty($dateTimeDia) ? DateUtil::adicionaIntervalo($dateTimeDia, 1, DateUtil::DIA) : null;
        
        // Trata caso o intervalo termine no dia sguinte e ainda deve ser considerado na apuração do dia corrente
        $idDiaOcorrenciaNovaSaida = self::DEFAULT_DIA_ID;
        if (($novaSaidaAlmoco > $tempoLimiteDiaAnterior) && (!is_null($tempoLimiteDiaAnterior))) {
            $idDiaOcorrenciaNovaSaida = self::IDDIAOCORRENCIA_FORCA_DIA_ANTERIOR;
        }
        
        // Insere marcação projetada
        $this->editMarcacao(
            $nrVinculo,
            $nrSaidaEntradaAlmoco,
            $dia,
            $novaSaidaAlmoco->format('H:i'),
            $idDiaOcorrenciaNovaSaida
        );
    }

    public function getMarcacaoSeguinte(
        int $nrVinculo,
        int $nrMarcacao
    ) {
        $nrOrg = (int) $this->environment->getNrOrg();

        $marcacao = $this->marcacaoRepository->getMarcacaoByNrmarcacaoponto(
            $nrOrg, $nrMarcacao
        );

        $date = (string) $marcacao['DIA'];

        $marcacoesDia = $this->getMarcacoesDia($nrVinculo, $date);

        $marcacaoSeguinte = false;
        $passouMarcacao = false;

        foreach ($marcacoesDia as $marc) {
            if ($passouMarcacao) {
                $marcacaoSeguinte = $marc;
                break;
            }

            if (strval($nrMarcacao) === $marc['NRMARCACAOPONTO']) {
                $passouMarcacao = true;
            }
        }

        return $marcacaoSeguinte;
    }

    public function removeMarcacoesDia(
        int    $nrVinculo,
        string $date
    ) {
        $nrOrg = (int) $this->environment->getNrOrg();

        $params = [
            'P_NRORG'       => $nrOrg,
            'P_NRVINCULO'   => $nrVinculo,
            'P_DATE'        => $date,
        ];

        $this->marcacaoRepository->removeMarcacoesDia($params);
    }

    public function getDetalheAcompanhamento(
        int $nrVinculo,
        array $dates
    ): array {
        $nrOrg = (int) $this->environment->getNrOrg();
        $datesSql = ArrayUtil::toStringSQL($dates);

        $params = [
            'P_NRORG'       => $nrOrg,
            'P_NRVINCULOM'  => $nrVinculo,
            'P_DATES'       => $datesSql,
        ];

        $marcacoes = $this->marcacaoRepository->getDetalheAcompanhamento($params);
        $marcacoes = $this->setIdColetorDetalheAcompanhamento($marcacoes);

        return $marcacoes;
    }

    private function setIdColetorDetalheAcompanhamento(
        array $marcacoes
    ): array {
        $paramOrg = $this->paramOrgService->get();

        if (!$paramOrg) {
            return $marcacoes;
        }

        $columnIdMap = [
            'NRCOLETORFUNC'         => 'COLETOR_FUNCIONARIO',
            'NRCOLETORGESTOR'       => 'COLETOR_GESTOR',
            'NRCOLETORMARCAUTOM'    => 'COLETOR_AUTOMATICO',
        ];

        foreach ($marcacoes as &$marc) {
            foreach ($columnIdMap as $column => $idName) {
                $idValue = false;
                $columnValue = $paramOrg[$column] ?? null;
                $sameColetor = $marc['NRCOLETOR'] === $columnValue;

                if (isset($columnValue) && $sameColetor) {
                    $idValue = true;
                }

                $marc[$idName] = $idValue;
            }
        }

        return $marcacoes;
    }
    
    public static function getInstance() {
        $instanceManager = InstanceManager::getInstance();
        return $instanceManager->getService('MarcacaoService');
    }

    public function getComprovanteReport(
        int $nrMarcacao
    ): array {
        $reportLink = $this->reportService->generateBirtReport(
            [
                'reportName' => 'comprovante_marcacao',
                'P_MARCACAO' => $nrMarcacao,
                'P_COMPETENCIA' => $this->environment->getDtmescomptencDiaCorrente(),
            ]
        );
        $reportFile = $this->reportService->downloadReport($reportLink);
        $reportFile['name'] = "comprovante_marcacao_${nrMarcacao}";
        return $reportFile;
    }
    
    public function createOcorAprovUltMarcacoes(
        int     $nrVinculo,
        string  $data,
        ?int    $nrMotivoJust = null,
        ?string $justificativa = null,
        ?string $anexoJustificativa = null
    ): void {
        if (empty($nrMotivoJust)) {
            throw new \Exception(Messages::MARCACAO_MANUAL_OBRIGATORIA);
        }
        
        foreach ($this->ultimasMarcacoesCriadas as $nrMarcacao) {
            $this->ocorrenciaService->setAprovacaoMarcacaoOcorrencia(
                $nrVinculo,
                $data,
                $nrMarcacao,
                $nrMotivoJust,
                $justificativa,
                $anexoJustificativa
            );
        }
    }
    
    public function validaRemocaoMarcacao(
        int $nrMarcacao
    ): void {
        if ($this->isMarcacaoPendenteAprovacao($nrMarcacao)) {
            throw new \Exception(Messages::MARCACAO_DELETE_FAIL);
        }
    }
    
    public function validaEdicaoMarcacao(
        int $nrMarcacao
    ): void {
        if ($this->isMarcacaoPendenteAprovacao($nrMarcacao) || $this->isMarcacaoNaoAprovada($nrMarcacao)) {
            throw new \Exception(Messages::MARCACAO_EDIT_FAIL);
        }
    }
    
    public function isMarcacaoNaoAprovada(
        int $nrMarcacao
    ): bool {
        $nrOrg = $this->environment->getNrorg();
        $marcacao = $this->entityManager->getRepository(Repositories::GPE_MARCACAOPONTO)->findOneBy([
            'nrorg' => $nrOrg, 'nrmarcacaoponto' => $nrMarcacao
        ]);
        
        if (empty($marcacao)) {
            return true;
        }
        
        if ($marcacao->getIdexigeaprovacao() == Utils::NAO) {
            return false;
        }
        
        $ocorrenciaAprovacao = $this->entityManager->getRepository(Repositories::PTO_PROGOCORRENCIA)->findOneBy([
            'nrorg' => $nrOrg, 'nrtipoprogocorrencia' => OcorrenciaService::NRTIPOPROGOCORRENCIA_AUTORIZARMARCACAO,
            'nrmarcacaoponto'=> $nrMarcacao, 'idativo' => Utils::SIM, 'idprogocorconfirmada' => Utils::SIM
        ]);
        
        return empty($ocorrenciaAprovacao);
    }
    
    public function isMarcacaoPendenteAprovacao(
        int $nrMarcacao
    ): bool {
        $nrOrg = $this->environment->getNrorg();
        $marcacao = $this->entityManager->getRepository(Repositories::GPE_MARCACAOPONTO)->findOneBy([
            'nrorg' => $nrOrg, 'nrmarcacaoponto' => $nrMarcacao
        ]);
        
        if (empty($marcacao) || $marcacao->getIdexigeaprovacao() == Utils::NAO) {
            return false;
        }
        
        $ocorrenciaAprovacao = $this->entityManager->getRepository(Repositories::PTO_PROGOCORRENCIA)->findOneBy([
            'nrorg' => $nrOrg, 'nrtipoprogocorrencia' => OcorrenciaService::NRTIPOPROGOCORRENCIA_AUTORIZARMARCACAO,
            'nrmarcacaoponto'=> $nrMarcacao, 'idativo' => Utils::SIM, 'idprogocorconfirmada' => 'P'
        ]);
        
        return !empty($ocorrenciaAprovacao);
    }
    
    public function isMomentoForaHorario(
        int $nrVinculo
    ): bool {
        if (!$this->paramOrgService->isConfirmacaoMarcNecessaria()) {
            return false;
        }

        if (!$this->environment->isRequestFromPortalFuncionario()) {
            return false;
        }

        return $this->isMomentoForaHorarioBase($nrVinculo);
    }

    public function isMomentoForaHorarioBase(
        int $nrVinculo
    ): bool {
        $dataAtual = DateUtil::getDataAtual();
        $horarioBase = $this->vinculoService->getHorarioBaseAtual($nrVinculo);
        
        $primeiraMarcacao = $this->horarioService->getPrimeiraMarcacaoHorario($horarioBase['NRHORDIAM'] ?? null);
        $ultimaMarcacao = $this->horarioService->getUltimaMarcacaoHorario($horarioBase['NRHORDIAM'] ?? null);
        if ($dataAtual >= $primeiraMarcacao && $dataAtual <= $ultimaMarcacao) {
            return false;
        }
        
        $horarioBaseAnterior = $this->vinculoService->getHorarioBaseAnterior($nrVinculo);
        $ultimaMarcacaoAnterior = $this->horarioService->getUltimaMarcacaoHorario(
            $horarioBaseAnterior['NRHORDIAM'] ?? null,
            DateUtil::subtraiIntervalo(DateUtil::truncateData($dataAtual), 1, DateUtil::DIA)
        );
        if ($dataAtual < $primeiraMarcacao && $dataAtual <= $ultimaMarcacaoAnterior) {
            return false;
        }
        
        $horarioBaseSeguinte = $this->vinculoService->getHorarioBaseSeguinte($nrVinculo);
        $primeiraMarcacaoSeguinte = $this->horarioService->getPrimeiraMarcacaoHorario(
            $horarioBaseSeguinte['NRHORDIAM'] ?? null,
            DateUtil::adicionaIntervalo(DateUtil::truncateData($dataAtual), 1, DateUtil::DIA)
        );
        if ($dataAtual > $primeiraMarcacao && $dataAtual > $primeiraMarcacaoSeguinte) {
            return false;
        }
        
        return true;
    }

    public function isColetorNotEditable(int $nrColetor): bool {
        $coletoresEditaveis = [
            $this->paramOrgService->getColetorPG(),
            $this->paramOrgService->getColetorAutomatica(),
            $this->paramOrgService->getColetorGeradaPeloSistema(),
            $this->paramOrgService->getColetorViradaDia(),
            $this->paramOrgService->getColetorProjetada(),
        ];

        $isColetorNotEditable = $this->environment->isRequestFromPortalGestor() &&
            !in_array($nrColetor, $coletoresEditaveis) &&
            !$this->operadorService->operadorIsDP($this->environment->getCdoperador()) &&
            $this->paramOrgService->isSetBloqueioPGGestorEdicaoMarc();

        return $isColetorNotEditable;
    }

    /**
     * Verifica se ja nao existe uma marcacao no mesmo minuto da hora informada
     */
    public function isNovaMaracaoDuplicante(
        int $nrVinculo,
        \DateTime $dataHora
    ): bool {
        return $this->entityManager->getRepository(Repositories::GPE_MARCACAOPONTO)->existeMarcacaoNoMesmoMinuto(
            $this->environment->getNrOrg(),
            $nrVinculo,
            $dataHora
        );
    }

    /**
     * Busca as marcacoes do dia tratando a busca convencional por marcacoes com a conversao para horarios apenas
     */
    public function getHorariosMarcacoesDia(
        int $nrVinculo,
        \DateTime $dataHora
    ): array {
        $marcacoesDia = $this->entityManager->getRepository(Repositories::GPE_MARCACAOPONTO)->getMarcacoesDiaExato(
            $this->environment->getNrOrg(),
            $nrVinculo,
            $dataHora
        );

        return array_map(
            function ($item) {
                return $item->getDthrajustada()->format(DateUtil::FORMATO_BRASILEIRO_HORA);
            },
            $marcacoesDia ?? []
        );
    }
    
    public static function marcacaoComDiaOcorrencia(
        \DateTime   $data,
        string      $diaOcoccorrencia
    ): \DateTime {
        switch ($diaOcoccorrencia) {
            case self::IDDIAOCORRENCIA_DIA_ANTERIOR:
            case self::IDDIAOCORRENCIA_FORCA_DIA_ANTERIOR:
                return DateUtil::subtraiIntervalo($data, 1, DateUtil::DIA);
            case self::IDDIAOCORRENCIA_DIA_SEGUINTE:
            case self::IDDIAOCORRENCIA_FORCA_DIA_SEGUINTE:
                return DateUtil::adicionaIntervalo($data, 1, DateUtil::DIA);
        }
        return $data;
    }
    
    public static function marcacaoComDiaOcorrenciaReal(
        \DateTime   $data,
        string      $diaOcoccorrencia
    ): \DateTime {
        switch ($diaOcoccorrencia) {
            case self::IDDIAOCORRENCIA_DIA_ANTERIOR:
            case self::IDDIAOCORRENCIA_FORCA_DIA_ANTERIOR:
                return DateUtil::adicionaIntervalo($data, 1, DateUtil::DIA);
            case self::IDDIAOCORRENCIA_DIA_SEGUINTE:
            case self::IDDIAOCORRENCIA_FORCA_DIA_SEGUINTE:
                return DateUtil::subtraiIntervalo($data, 1, DateUtil::DIA);
        }
        return $data;
    }

    /**
     * De acordo com os intervalos cadastrados para o horario previsto para o dia da batida, decide a sinalizacao da marcacao
     *
     * @return string (IMPORTACAO_DIA_CORRENTE, IMPORTACAO_DIA_SEGUINTE, IMPORTACAO_DIA_ANTERIOR)
     */
    public function getIdDiaMarcFromHora(string $nrVinculo, \DateTime $dtHrMarc): string {
        /* Deteccao de virada para o dia anterior */
        
        $nrHordiamVinculoDiaAnterior = $this->getNrHordiamVinculoDiaAnterior($nrVinculo, $dtHrMarc);
        
        if ($this->isMarcImportadaDiaAnterior($dtHrMarc, $nrHordiamVinculoDiaAnterior)) {
            return self::IMPORTACAO_DIA_ANTERIOR;
        }
        
        /* Deteccao de virada para o dia seguinte */
        
        $nrHordiamVinculoDiaSeguinte = $this->getNrHordiamVinculoDiaSeguinte($nrVinculo, $dtHrMarc);
        
        if ($this->isMarcImportadaDiaSeguinte($dtHrMarc, $nrHordiamVinculoDiaSeguinte)) {
            return self::IMPORTACAO_DIA_SEGUINTE;
        }
        
        /* Dia convencional */
        
        return self::IMPORTACAO_DEFAULT_DIA;
    }
    
    /**
     * Busca o horario do dia imediatamente anterior ao da batida
     * 
     * @param string    $nrVinculo
     * @param \DateTime $dtHrMarc
     * 
     * @return ?string
     */
    private function getNrHordiamVinculoDiaAnterior(string $nrVinculo, \DateTime $dtHrMarc): ?string {
        $dadosHorarioVinculo = $this->marcacaoRepository->getDadosHorarioVinculo(
            $nrVinculo,
            $this->environment->getNrorg(),
            DateUtil::subtraiIntervalo($dtHrMarc, 1)->format(DateUtil::FORMATO_DATA)
        );
        
        return $dadosHorarioVinculo['NRHORDIAM'] ?? null;
    }
    
    /**
     * Busca o horario do dia imediatamente seguinte ao da batida
     * 
     * @param string    $nrVinculo
     * @param \DateTime $dtHrMarc
     * 
     * @return ?string
     */
    private function getNrHordiamVinculoDiaSeguinte(string $nrVinculo, \DateTime $dtHrMarc): ?string {
        $dadosHorarioVinculo = $this->marcacaoRepository->getDadosHorarioVinculo(
            $nrVinculo,
            $this->environment->getNrorg(),
            DateUtil::adicionaIntervalo($dtHrMarc, 1)->format(DateUtil::FORMATO_DATA)
        );
        
        return $dadosHorarioVinculo['NRHORDIAM'] ?? null;
    }
    
    /**
     * Verifica se a marcacao sera sinalizada como anterior considerando o horario do dia anterior
     * 
     * @param \DateTime $dtHrMarc
     * @param ?string   $nrHordiam
     * 
     * @return ?string
     */
    private function isMarcImportadaDiaAnterior(\DateTime $dtHrMarc, ?string $nrHordiam): bool {
        $naoExisteViradaNoHorario = !$this->horarioService->horarioIncluiDiaAnterior($this->environment->getNrorg(), $nrHordiam);
            
        $horarioNaoInformado = empty($nrHordiam);
        
        if ($naoExisteViradaNoHorario || $horarioNaoInformado) {
            return false;
        }
        
        return $this->isMarcImportadaDentroToleranciaDiaAnterior($dtHrMarc, $nrHordiam);
    }
    
    /**
     * Verifica se a marcacao sera sinalizada como seguinte considerando o horario do dia seguinte
     * 
     * @param \DateTime $dtHrMarc
     * @param ?string   $nrHordiam
     * 
     * @return ?string
     */
    private function isMarcImportadaDiaSeguinte(\DateTime $dtHrMarc, ?string $nrHordiam): bool {
        $naoExisteViradaNoHorario = !$this->horarioService->horarioIncluiDiaSeguinte($this->environment->getNrorg(), $nrHordiam);
            
        $horarioNaoInformado = empty($nrHordiam);
        
        if ($naoExisteViradaNoHorario || $horarioNaoInformado) {
            return false;
        }
        
        return $this->isMarcImportadaDentroToleranciaDiaSeguinte($dtHrMarc, $nrHordiam);
    }
    
    /**
     * Verifica se a marcacao esta dentro do periodo previsto para ser sinalizada como anterior, considerando um tempo definido de tolerancia a partir da ultima batida do horario
     * 
     * @param \DateTime $dtHrMarc
     * @param string    $nrHordiam
     * 
     * @return bool
     */
    private function isMarcImportadaDentroToleranciaDiaAnterior(\DateTime $dtHrMarc, string $nrHordiam): bool {
        $ultimaMarcacaoDia = $this->horarioService->getUltimaHoraHorario($nrHordiam, DateUtil::truncateData($dtHrMarc));
        
        $toleranciaDiaAnterior = DateUtil::adicionaIntervalo(
            $ultimaMarcacaoDia,
            DateUtil::getHoraFloatDeString(self::IMPORTACAO_TEMPO_TOLERANCIA_ANTERIOR),
            DateUtil::HORAS
        );
        
        return $dtHrMarc <= $toleranciaDiaAnterior;
    }
    
    /**
     * Verifica se a marcacao esta dentro do periodo previsto para ser sinalizada como seguinte, considerando um tempo definido de tolerancia a partir da primeira batida do horario
     * 
     * @param \DateTime $dtHrMarc
     * @param string    $nrHordiam
     * 
     * @return bool
     */
    private function isMarcImportadaDentroToleranciaDiaSeguinte(\DateTime $dtHrMarc, string $nrHordiam): bool {
        $primeiraMarcacaoDia = $this->horarioService->getPrimeiraHoraHorario($nrHordiam, DateUtil::truncateData($dtHrMarc));
        
        $toleranciaDiaSeguinte = DateUtil::subtraiIntervalo(
            $primeiraMarcacaoDia,
            DateUtil::getHoraFloatDeString(self::IMPORTACAO_TEMPO_TOLERANCIA_SEGUINTE),
            DateUtil::HORAS
        );
        
        /* Caso o horario seja somente no dia seguinte ou no dia anterior */
        
        if (!$this->horarioService->horarioIncluiDiaCorrente($this->environment->getNrorg(), $nrHordiam)) {
            $toleranciaDiaSeguinte = DateUtil::adicionaIntervalo(
                $toleranciaDiaSeguinte,
                1,
                DateUtil::DIA
            );
        }
        
        return $dtHrMarc >= $toleranciaDiaSeguinte;
    }
    public function verifyMarcacaoIsDuplicada(
        int     $nrVinculo,
        string  $date,
        ?string $hrEntrada,
        ?string $hrSaida
    ) {
        $entradaDuplicada = false;
        $saidaDuplicada = false;
        
        if($hrEntrada) {
            $dataEntrada = $date . ' ' . $hrEntrada;

            $entradaDuplicada = $this->isNovaMaracaoDuplicante($nrVinculo, \DateTime::createFromFormat('d/m/Y H:i', $dataEntrada));
        }
        
        if($hrSaida) {
            $dataSaida = $date . ' ' . $hrSaida;

            $saidaDuplicada = $this->isNovaMaracaoDuplicante($nrVinculo, \DateTime::createFromFormat('d/m/Y H:i', $dataSaida));
        }
        
        $dataDuplicada = $entradaDuplicada ? $dataEntrada : ($saidaDuplicada ? $dataSaida : false);

        if ($dataDuplicada) {
            throw new \Exception("Marcação: '{$dataDuplicada}' não criada pois já existe uma marcação com mesma data e horário.");
        }
    }
}
